// (C) 2017-2020 by folkert van heusden, released under AGPL v3.0
#include "config.h"
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>

#include "target_extpipe.h"
#include "error.h"
#include "exec.h"
#include "log.h"
#include "picio.h"
#include "utils.h"
#include "formats.h"

target_extpipe::target_extpipe(const std::string & id, const std::string & descr, source *const s, const std::string & store_path, const std::string & prefix, const std::string & fmt, const int quality, const int max_time, const double interval, const std::vector<filter *> *const filters, const std::string & exec_start, const std::string & exec_cycle, const std::string & exec_end, const std::string & cmd, configuration_t *const cfg, const bool is_view_proxy, const bool use_previous_when_fail) : target(id, descr, s, store_path, prefix, fmt, max_time, interval, filters, exec_start, exec_cycle, exec_end, -1.0, cfg, is_view_proxy, use_previous_when_fail), quality(quality), cmd(cmd)
{
	if (this -> descr == "")
		this -> descr = store_path + "/" + prefix;
}

target_extpipe::~target_extpipe()
{
	stop();
}

static void put_frame(FILE *fh, const uint8_t *const in, const int w, const int h)
{
	size_t n = w * h + w * h / 2;
	uint8_t *temp = (uint8_t *)malloc(n);

	Bitmap2Yuv420p_calc2(temp, in, w, h);

	fwrite(temp, 1, n, fh);

	free(temp);
}

void target_extpipe::operator()()
{
	set_thread_name("storee_" + prefix);

	s -> register_user();

	time_t cut_ts = time(nullptr) + max_time;

	uint64_t prev_ts = 0;
	bool is_start = true;
	std::string name;
	unsigned f_nr = 0;

	FILE *p_fd = nullptr;

	uint8_t *prev_frame = nullptr;

	bool first = true;
	std::string workCmd = cmd;

	for(;!local_stop_flag;) {
		pauseCheck();

		double before_ts = get_ts();

		int w = -1, h = -1;
		uint8_t *work = nullptr;
		size_t work_len = 0;
		if (s -> get_frame(E_RGB, quality, &prev_ts, &w, &h, &work, &work_len, use_previous_when_fail)) {
			if (max_time > 0 && time(nullptr) >= cut_ts) {
				log(id, LL_DEBUG, "new file");

				if (p_fd) {
					pclose(p_fd);
					p_fd = nullptr;
				}

				cut_ts = time(nullptr) + max_time;
			}

			if (p_fd == nullptr) {
				name = gen_filename(s, fmt, store_path, prefix, "", get_us(), f_nr++);
				create_path(name);
				register_file(name);

				if (!exec_start.empty() && is_start) {
					exec(exec_start, name);
					is_start = false;
				}
				else if (!exec_cycle.empty()) {
					exec(exec_cycle, name);
				}

				int fps = interval <= 0 ? 25 : std::max(1, int(1.0 / interval));

				if (first) {
					first = false;

					char fps_str[16];
					snprintf(fps_str, sizeof fps_str, "%d", fps);

					workCmd = search_replace(workCmd, "%fps", fps_str);

					char w_str[16], h_str[16];
					snprintf(w_str, sizeof w_str, "%d", w);
					snprintf(h_str, sizeof h_str, "%d", h);
					workCmd = search_replace(workCmd, "%w", w_str);
					workCmd = search_replace(workCmd, "%h", h_str);

					workCmd = search_replace(workCmd, "%f", name);

					log(id, LL_DEBUG, "Will invoke %s", workCmd.c_str());
				}

				p_fd = exec(workCmd);

				if (pre_record) {
					log(id, LL_DEBUG_VERBOSE, "Write pre-recorded frames");

					for(frame_t pair : *pre_record) {
						if (pair.e == E_JPEG) {
							unsigned char *temp = nullptr;
							int dw = -1, dh = -1;
							if (!read_JPEG_memory(work, work_len, &dw, &dh, &temp))
								log(id, LL_INFO, "JPEG decode error");
							else {
								put_frame(p_fd, temp, dw, dh);
								free(temp);
							}
						}
						else {
							put_frame(p_fd, pair.data, w, h);
						}

						free(pair.data);
					}

					delete pre_record;
					pre_record = nullptr;
				}
			}

			log(id, LL_DEBUG_VERBOSE, "Write frame");
			source *cur_s = is_view_proxy ? ((view *)s) -> get_current_source() : s;
			instance_t *inst = find_instance_by_interface(cfg, cur_s);
			apply_filters(inst, cur_s, filters, prev_frame, work, prev_ts, w, h);

			put_frame(p_fd, work, w, h);

			free(prev_frame);
			prev_frame = work;
		}

		handle_fps(before_ts);
	}

	if (p_fd != nullptr)
		pclose(p_fd);

	free(prev_frame);

	if (!exec_end.empty())
		exec(exec_end, name);

	s -> unregister_user();
}
